import os

import boto3

from awstin.config import aws_config
from awstin.constants import TEST_DYNAMODB_ENDPOINT

# Testing parameter to change table listing page size
_PAGE_SIZE = 100


class DynamoDB:
    """
    A client for typical use of DynamoDB.
    """
    def __init__(self, timeout=5.0, max_retries=3):
        """
        Parameters
        ----------
        timeout : float, optional
            Timeout for establishing a conneciton to DynamoDB (default 5.0)
        max_retries : int, optional
            Max retries for establishing a connection to DynamoDB (default 3)

        Raises
        ------
        EnvironmentError
            If neither the TEST_DYNAMODB_ENDPOINT or AWS_REGION environment
            variables are set
        """
        self.timeout = timeout
        self.max_retries = max_retries

        test_endpoint = os.environ.get(TEST_DYNAMODB_ENDPOINT)
        self.config = aws_config(
            timeout=timeout,
            max_retries=max_retries,
            endpoint=test_endpoint,
        )
        self.client = boto3.client('dynamodb', **self.config)
        self.resource = boto3.resource('dynamodb', **self.config)

    def list_tables(self):
        """
        Return a list of all table names in this DynamoDB instance.

        Returns
        -------
        list of str
            Table names
        """
        response = self.client.list_tables(Limit=_PAGE_SIZE)
        tables = response["TableNames"]

        while "LastEvaluatedTableName" in response:
            response = self.client.list_tables(
                Limit=_PAGE_SIZE,
                ExclusiveStartTableName=response["LastEvaluatedTableName"]
            )
            tables.extend(response["TableNames"])

        return tables

    def __getitem__(self, key):
        """
        Indexed access to DynamoDB tables.

        Returns
        -------
        Table
            the dynamodb table, if it exists.
        """
        return Table(self, key)


class Table:
    """
    Interface to a DynamoDB table.

    Items can be retrieved in a dict-like way:
    ```
    Table("table_name")[{"HashKeyName": "hashval", "SortKeyName": 123}]
    ```

    Items can also be retrieved from the table by a shorthand depending on the
    primary key. If it's only a partition key, items can be retrieved by the
    value of the partition key:
    ```
    Table("table_name")["hashval"]
    ```
    If it's a partition and sort key, items can be retrived by a hashkey,
    sortkey tuple:
    ```
    Table("table_name")[("hashval", 123)]
    ```
    """
    def __init__(self, dynamodb_client, table_name):
        """
        Paramters
        ---------
        client : DynamoDB
            DynamoDB client
        table_name : DynamoDB Table
            Table to convert into an awstin Table
        """
        self.name = table_name

        self._dynamodb = dynamodb_client
        self._boto3_table = dynamodb_client.resource.Table(table_name)

    def _get_primary_key(self, key):
        if isinstance(key, dict):
            primary_key = key
        else:
            table_description = self._dynamodb.client.describe_table(
                TableName=self.name,
            )
            partition_key, = [
                entry["AttributeName"]
                for entry in table_description["Table"]["KeySchema"]
                if entry["KeyType"] == "HASH"
            ]
            if isinstance(key, tuple):
                sort_key, = [
                    entry["AttributeName"]
                    for entry in table_description["Table"]["KeySchema"]
                    if entry["KeyType"] == "RANGE"
                ]
                primary_key = {partition_key: key[0], sort_key: key[1]}
            else:
                primary_key = {partition_key: key}
        return primary_key

    def __getitem__(self, key):
        """
        Get an item, given either a primary key as a dict, or given simply the
        value of the partition key if there is no sort key
        """
        primary_key = self._get_primary_key(key)
        return self._boto3_table.get_item(Key=primary_key)["Item"]

    def put_item(self, item):
        """
        Direct exposure of put_item
        """
        return self._boto3_table.put_item(Item=item)

    def update_item(self, *args, **kwargs):
        """
        For now, direct exposure of update_item
        """
        return self._boto3_table.update_item(*args, **kwargs)

    def delete_item(self, key):
        """
        Delete an item, given either a primary key as a dict, or given simply
        the value of the partition key if there is no sort key
        """
        primary_key = self._get_primary_key(key)
        self._boto3_table.delete_item(Key=primary_key)

    def scan(self):
        """
        Generate all items in the table, one at a time.

        Lazily queries for more items if needed.

        Yields
        ------
        dict
            an item in the table
        """
        # TODO: scan filters

        results = self._boto3_table.scan()
        items = results["Items"]
        yield from items

        while "LastEvaluatedKey" in results:
            results = self._boto3_table.scan(
                ExclusiveStartKey=results["LastEvaluatedKey"],
            )
            items = results["Items"]
            yield from items

    # TODO: Batch Update



import unittest

from awstin.apigateway import auth
from awstin.awslambda import lambda_handler


def auth_event_parser(event, _context):
    token = event["headers"]["AuthToken"]
    resource_arn = event["methodArn"]
    principal_id = event["requestContext"]["connectionId"]

    return token, resource_arn, principal_id


@lambda_handler(auth_event_parser)
def token_auth(token, resource_arn, principal_id):
    if token == "good token":
        return auth.accept(principal_id, resource_arn)
    elif token == "bad token":
        return auth.reject(principal_id, resource_arn)
    elif token == "unauthorized token":
        return auth.unauthorized()
    else:
        return auth.invalid()


class TestAuthHandler(unittest.TestCase):
    def test_auth_handler_accept(self):
        event = dict(
            headers={"AuthToken": "good token"},
            methodArn="arn123",
            requestContext={"connectionId": "id456"},
        )
        context = {}

        # Invoke the lambda
        result = token_auth(event, context)

        expected_result = {
            "principalId": "id456",
            "policyDocument": {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Action": [
                            "execute-api:Invoke",
                        ],
                        "Effect": "Allow",
                        "Resource": ["arn123"],
                    }
                ]
            },
        }
        self.assertEqual(result, expected_result)

    def test_auth_handler_reject(self):
        event = dict(
            headers={"AuthToken": "bad token"},
            methodArn="arn123",
            requestContext={"connectionId": "id456"},
        )
        context = {}

        # Invoke the lambda
        result = token_auth(event, context)

        expected_result = {
            "principalId": "id456",
            "policyDocument": {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Action": [
                            "execute-api:Invoke",
                        ],
                        "Effect": "Deny",
                        "Resource": ["arn123"],
                    }
                ]
            },
        }
        self.assertEqual(result, expected_result)

    def test_auth_handler_unauthorized(self):
        event = dict(
            headers={"AuthToken": "unauthorized token"},
            methodArn="arn123",
            requestContext={"connectionId": "id456"},
        )
        context = {}

        # Invoke the lambda
        result = token_auth(event, context)

        expected_result = {
            'statusCode': 401,
            'body': 'Unauthorized',
        }
        self.assertEqual(result, expected_result)

    def test_auth_handler_invalid_other_value(self):
        event = dict(
            headers={"AuthToken": "unexpected value"},
            methodArn="arn123",
            requestContext={"connectionId": "id456"},
        )
        context = {}

        # Invoke the lambda
        result = token_auth(event, context)

        expected_result = {
            'statusCode': 500,
            'body': 'Invalid',
        }
        self.assertEqual(result, expected_result)



import os

from botocore.client import Config

from awstin.constants import AWS_REGION


def aws_config(timeout=5.0, max_retries=3, endpoint=None):
    """
    Create kwargs used to configure a boto3 client or resource.

    Parameters
    ----------
    timeout : float, optional
        Timeout for establishing a conneciton to DynamoDB (default 5.0)
    max_retries : int, optional
        Max retries for establishing a connection to DynamoDB (default 3)
    endpoint : str or None, optional
        Endpoint for the AWS service for testing. If not provided, the
        AWS_REGION environment variable will be used to specify the AWS region

    Returns
    -------
    dict
        kwargs to pass to boto3 client or resource

    Raises
    ------
    EnvironmentError
        If the AWS_REGION environment variable is required but is not set
    """
    region_name = os.environ.get(AWS_REGION)
    if endpoint:
        kwargs = {"endpoint_url": endpoint}
    elif region_name:
        kwargs = {"region_name": region_name}
    else:
        raise EnvironmentError(
            "Please set either the AWS_REGION environment variable or set a "
            "testing endpoint."
        )

    config = Config(
        connect_timeout=timeout,
        retries={'max_attempts': max_retries},
    )

    return {
        "config": config,
        **kwargs
    }




def accept(principal_id, resource_arn):
    """
    Return an auth lambda response granting access to the given resource ARN
    to the given principle ID.

    Parameters
    ----------
    principal_id : str
        The principal ID to grant access to
    resource_arn : str
        The ARN of the resource to grant access to

    Returns
    -------
    dict
        Auth lambda response
    """
    return _auth_response(principal_id, resource_arn, "Allow")


def reject(principal_id, resource_arn):
    """
    Return an auth lambda response rejecting access to the given resource ARN
    to the given principle ID.

    Parameters
    ----------
    principal_id : str
        The principal ID to reject access to
    resource_arn : str
        The ARN of the resource to reject access to

    Returns
    -------
    dict
        Auth lambda response
    """
    return _auth_response(principal_id, resource_arn, "Deny")


def unauthorized(body='Unauthorized'):
    """
    Return an auth lambda response indicating the requester is unauthorized.

    Parameters
    ----------
    body : str, optional
        Optional resposnse body. Default "Unauthorized"

    Returns
    -------
    dict
        Auth lambda response
    """
    return {
        'statusCode': 401,
        'body': body,
    }


def invalid(body='Invalid'):
    """
    Return an auth lambda response indicating the request is invalid.

    Parameters
    ----------
    body : str, optional
        Optional resposnse body. Default "Invalid"

    Returns
    -------
    dict
        Auth lambda response
    """
    return {
        'statusCode': 500,
        'body': body,
    }


def _auth_response(principal_id, resource_arn, effect):
    return {
        "principalId": principal_id,
        "policyDocument": {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Action": [
                        "execute-api:Invoke",
                    ],
                    "Effect": effect,
                    "Resource": [resource_arn],
                }
            ]
        },
    }



import os
import unittest

from awstin.environment import set_env

UNLIKELY_ENV_VAR = "THIS_ENV_VAR_IS_UNLIKELY_TO_BE_IN_USE_150y71i1"


class TestSetEnv(unittest.TestCase):

    def setUp(self):
        # Always remove the used environment variable
        def reset_env_var():
            if UNLIKELY_ENV_VAR in os.environ:
                del os.environ[UNLIKELY_ENV_VAR]
        self.addCleanup(reset_env_var)

    def test_temp_set_new_env_var(self):
        self.assertNotIn(UNLIKELY_ENV_VAR, os.environ)

        temp_env = {UNLIKELY_ENV_VAR: "test value"}

        with set_env(**temp_env):
            self.assertEqual(os.environ.get(UNLIKELY_ENV_VAR), "test value")

        self.assertNotIn(UNLIKELY_ENV_VAR, os.environ)

    def test_env_var_already_in_use(self):
        os.environ[UNLIKELY_ENV_VAR] = "old value"

        temp_env = {UNLIKELY_ENV_VAR: "new value"}

        with set_env(**temp_env):
            self.assertEqual(os.environ.get(UNLIKELY_ENV_VAR), "new value")

        self.assertEqual(os.environ.get(UNLIKELY_ENV_VAR), "old value")

    def test_remove_env_var(self):
        os.environ[UNLIKELY_ENV_VAR] = "old value"

        temp_env = {UNLIKELY_ENV_VAR: None}

        with set_env(**temp_env):
            self.assertNotIn(UNLIKELY_ENV_VAR, os.environ)

        self.assertEqual(os.environ.get(UNLIKELY_ENV_VAR), "old value")

    def test_remove_nonexist_env_var(self):
        # Test setting an env var to None that wasn't previously set
        temp_env = {UNLIKELY_ENV_VAR: None}

        with set_env(**temp_env):
            self.assertNotIn(UNLIKELY_ENV_VAR, os.environ)

        self.assertNotIn(UNLIKELY_ENV_VAR, os.environ)



import unittest
from unittest import mock

from awstin.config import __name__ as CONFIG_NAME
from awstin.config import aws_config
from awstin.constants import AWS_REGION
from awstin.environment import set_env


class TestConfig(unittest.TestCase):
    def test_aws_config_production(self):
        # Check that in a production environment, the resource is called
        # with the right kwargs
        temp_env = {
            AWS_REGION: "SOME REGION",
        }

        with set_env(**temp_env):
            config = aws_config("test_table_name")

        self.assertEqual(config["region_name"], "SOME REGION")
        self.assertNotIn("endpoint_url", config)

    def test_aws_config_no_env_vars(self):
        temp_env = {
            AWS_REGION: None,
        }

        with set_env(**temp_env):
            with self.assertRaises(EnvironmentError) as err:
                aws_config("a table name")

        self.assertEqual(
            "Please set either the AWS_REGION environment variable or set a "
            "testing endpoint.",
            str(err.exception)
        )

    def test_aws_config_kwargs(self):
        # Check that kwargs are passed to the resource properly

        with mock.patch(CONFIG_NAME+".Config") as mock_config:
            aws_config(timeout=11, max_retries=55, endpoint="123.456")

        mock_config.assert_called_once_with(
            connect_timeout=11,
            retries={"max_attempts": 55}
        )



import logging

logger = logging.getLogger(__name__)


def lambda_handler(event_parser):
    """
    Decorator factory for wrapping a lambda handler in a boilerplate event
    logger and parser.

    Parameters
    ----------
    event_parser : callable
        Parser of a lambda handler input. Should take as an input the event and
        context as (dict, dict), and return a list of arguments that the
        wrapped handler should use.

    Returns
    -------
    handler : callable
        Decorator for the Lambda handler, accepting a LambdaEvent. It logs
        the raw incoming event and the result.
    """

    def handler(func):
        def wrapped(event, context):
            logger.info(f"Event: {event!r}")
            args = event_parser(event, context)
            if isinstance(args, tuple):
                result = func(*args)
            elif isinstance(args, dict):
                result = func(**args)
            else:
                result = func(args)
            logger.info(f"Result: {result!r}")
            return result
        return wrapped
    return handler



import boto3


class Websocket:
    """
    Serverless-to-client push via websocket
    """
    def __init__(self, domain_name, stage=None):
        """
        Parameters
        ----------
        connection_id : str
            Callback URL of the websocket
        """
        endpoint_url = f"https://{domain_name}"
        if stage:
            endpoint_url += f"/{stage}"
        self.api_client = boto3.client(
            'apigatewaymanagementapi',
            endpoint_url=endpoint_url,
        )

    def send(self, connection_id, message):
        """
        Send a message to the user
        """
        self.api_client.post_to_connection(
            Data=message,
            ConnectionId=connection_id,
        )



# -- AWS Lambda environment variables

#: Lambda function handler
_HANDLER = "_HANDLER"

#: Region
AWS_REGION = "AWS_REGION"

#: Runtime ID
AWS_EXECUTION_ENV = "AWS_EXECUTION_ENV"

#: Lambda name
AWS_LAMBDA_FUNCTION_NAME = "AWS_LAMBDA_FUNCTION_NAME"

#: Lambda memory parameter
AWS_LAMBDA_FUNCTION_MEMORY_SIZE = "AWS_LAMBDA_FUNCTION_MEMORY_SIZE"

#: Lambda version
AWS_LAMBDA_FUNCTION_VERSION = "AWS_LAMBDA_FUNCTION_VERSION"

#: Log group name
AWS_LAMBDA_LOG_GROUP_NAME = "AWS_LAMBDA_LOG_GROUP_NAME"

#: Log stream name
AWS_LAMBDA_LOG_STREAM_NAME = "AWS_LAMBDA_LOG_STREAM_NAME"

#: AWS access key (from execution role)
AWS_ACCESS_KEY_ID = "AWS_ACCESS_KEY_ID"

#: AWS secret key (from execution role)
AWS_SECRET_ACCESS_KEY = "AWS_SECRET_ACCESS_KEY"

#: Session token (from execution role)
AWS_SESSION_TOKEN = "AWS_SESSION_TOKEN"

#: Runtime API: None unless using a custom runtime
AWS_LAMBDA_RUNTIME_API = "AWS_LAMBDA_RUNTIME_API"

#: Path to the lambda
LAMBDA_TASK_ROOT = "LAMBDA_TASK_ROOT"

#: Path to runtime libraries
LAMBDA_RUNTIME_DIR = "LAMBDA_RUNTIME_DIR"

#: Timezone
TZ = "TZ"


# -- Testing Environment Variables

#: DynamodDB service endpoint, for example if running a DynamoDB container
TEST_DYNAMODB_ENDPOINT = "TEST_DYNAMODB_ENDPOINT"

#: SNS service endpoint for testing
TEST_SNS_ENDPOINT = "TEST_SNS_ENDPOINT"



import contextlib
import warnings

from awstin.dynamodb import DynamoDB


@contextlib.contextmanager
def temporary_dynamodb_table(
        table_name,
        hashkey_name,
        hashkey_type="S",
        sortkey_name=None,
        sortkey_type="S",
        delay=5.0,
        max_attempts=10,):
    """
    Context manager creating a temporary DynamoDB table for testing.

    Parameters
    ----------
    table_name : str
        Name of the table to create
    hashkey_name : str
        Name of the hash key of the table
    hashkey_type : str, optional
        Type of the hash key ("S", "N", or "B"). Default "S"
    sortkey_name : str, optional
        Optional sort key for the temporary table
    sortkey_type : str, optional
        Type of the sort key if there is one ("S", "N", or "B"). Default "S"
    delay : float, optional
        Delay in seconds between checks if the table exists
    max_attempts : int, optional
        Max number of attempts to check if the table exists, after which the
        client gives up.
    """
    # TODO: make filter more specific
    warnings.simplefilter("ignore", ResourceWarning)

    dynamodb = DynamoDB()

    attribute_definitions = [
        {
            'AttributeName': hashkey_name,
            'AttributeType': hashkey_type,
        },
    ]
    key_schema = [
        {
            'AttributeName': hashkey_name,
            'KeyType': 'HASH',
        },
    ]
    if sortkey_name is not None and sortkey_type is not None:
        attribute_definitions.append(
            {
                'AttributeName': sortkey_name,
                'AttributeType': sortkey_type,
            }
        )
        key_schema.append(
            {
                'AttributeName': sortkey_name,
                'KeyType': 'RANGE',
            },
        )

    dynamodb.client.create_table(
        TableName=table_name,
        AttributeDefinitions=attribute_definitions,
        KeySchema=key_schema,
        ProvisionedThroughput={
            'ReadCapacityUnits': 123,
            'WriteCapacityUnits': 123
        },
    )

    exists_waiter = dynamodb.client.get_waiter("table_exists")
    not_exists_waiter = dynamodb.client.get_waiter("table_not_exists")

    result = exists_waiter.wait(
        TableName=table_name,
        WaiterConfig={
            'Delay': delay,
            'MaxAttempts': max_attempts,
        }
    )
    if result is not None:
        raise RuntimeError("Could not create table {!r}".format(table_name))

    try:
        yield dynamodb[table_name]
    finally:
        dynamodb.client.delete_table(TableName=table_name)

        result = not_exists_waiter.wait(
            TableName=table_name,
            WaiterConfig={
                'Delay': delay,
                'MaxAttempts': max_attempts,
            }
        )
        if result is not None:
            msg = "Could not delete table {!r}"
            raise RuntimeError(msg.format(table_name))



import unittest
import unittest.mock as mock

from awstin.dynamodb import DynamoDB
from awstin.dynamodb import __name__ as DYNAMODB_NAME
from awstin.dynamodb.testing import temporary_dynamodb_table


class TestTemporaryDynamoDBTable(unittest.TestCase):

    def test_create_dynamodb_table(self):
        dynamodb = DynamoDB()

        self.assertNotIn(
            "test_table_name",
            dynamodb.list_tables()
        )

        with temporary_dynamodb_table("test_table_name", "test_hashkey_name"):
            self.assertIn(
                "test_table_name",
                dynamodb.list_tables()
            )

        self.assertNotIn(
            "test_table_name",
            dynamodb.list_tables()
        )

    def test_create_dynamodb_table_composite_key(self):
        dynamodb = DynamoDB()

        with temporary_dynamodb_table(
            "test_table_name",
            "test_hashkey_name",
            sortkey_name="sortkey_name",
        ):
            self.assertIn(
                "test_table_name",
                dynamodb.list_tables()
            )
            table_desc = dynamodb.client.describe_table(
                TableName="test_table_name",
            )

        # Both a hash and sort key
        self.assertEqual(len(table_desc["Table"]["KeySchema"]), 2)

    def test_create_dynamodb_table_fails(self):
        fake_client = mock.Mock()

        # Mock the dynamodb client's list_tables method to return an empty
        # table names list, which should cause the loop to time out
        def fake_list_tables(): return {"TableNames": []}
        fake_client.list_tables = fake_list_tables

        mock_dynamodb = mock.patch(
            DYNAMODB_NAME + ".boto3.client",
            return_value=fake_client,
        )

        temp_table_ctx = temporary_dynamodb_table(
            "test_table_name",
            "test_hashkey_name",
            delay=0.1,
            max_attempts=1,
        )

        with mock_dynamodb:
            with self.assertRaises(RuntimeError) as err:
                with temp_table_ctx:
                    # Cannot create a table so we'll have an exception
                    pass

        self.assertEqual(
            "Could not create table 'test_table_name'",
            str(err.exception)
        )



import contextlib
import os


@contextlib.contextmanager
def set_env(**env_vars):
    """
    Context manager to temporarily set environment variables. Restores previous
    values on leaving

    Parameters
    ----------
    **env_vars : Dict(str, str or None)
        Environment variables to set temporarily. If None, the environment
        variable will be tempoarily removed.
    """
    old_environ = dict(os.environ)
    for key, value in env_vars.items():
        if value is None:
            if key in os.environ:
                del os.environ[key]
        else:
            os.environ[key] = value

    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)



import os
import unittest
from unittest import mock

import boto3

from awstin.config import aws_config
from awstin.constants import TEST_SNS_ENDPOINT
from awstin.sns import SNSTopic


class TestSNSTopic(unittest.TestCase):
    def setUp(self):
        config = aws_config(endpoint=os.environ.get(TEST_SNS_ENDPOINT))
        self.sns_client = boto3.client("sns", **config)

        self.addCleanup(self.cleanup_topics)

    def cleanup_topics(self):
        topics = self.sns_client.list_topics()["Topics"]
        arns = [topic["TopicArn"] for topic in topics]
        for arn in arns:
            self.sns_client.delete_topic(TopicArn=arn)

    def test_instantiate_creates_topic(self):
        SNSTopic("name_of_topic")
        topics = self.sns_client.list_topics()["Topics"]
        self.assertEqual(len(topics), 1)
        topic, = topics
        self.assertTrue(topic["TopicArn"].endswith("name_of_topic"))

    def test_post_to_sns_topic(self):
        # Mock out SNS publish - may be possible to integration test later
        # with subscriptions to lambdas.
        topic = SNSTopic("some_topic")

        mock_sns = mock.Mock()
        mock_sns.publish = mock.Mock(return_value={"MessageId": "msg_id"})

        with mock.patch.object(topic, "topic", mock_sns):
            message_id = topic.publish("a cool message")

        mock_sns.publish.assert_called_once_with(
            Message="a cool message",
            MessageAttributes={},
        )
        self.assertEqual(message_id, "msg_id")

    def test_post_to_sns_topic_attributes(self):
        topic = SNSTopic("atopic")

        # Create a phony subsscription for boto3 reasons
        topic.topic.subscribe(Protocol="http", Endpoint="http://0")

        with mock.patch.object(topic, "topic", wraps=topic.topic) as w_topic:
            topic.publish(
                "a_cool_message",
                attrib_a="a string",
                attrib_b=1234,
                attrib_c=["a", "b", False, None],
                attrib_d=b"bytes value",
            )

        w_topic.publish.assert_called_once_with(
            Message="a_cool_message",
            MessageAttributes={
                "attrib_a": {
                    "DataType": "String",
                    "StringValue": "a string",
                },
                "attrib_b": {
                    "DataType": "Number",
                    "StringValue": "1234",
                },
                "attrib_c": {
                    "DataType": "String.Array",
                    "StringValue": '["a", "b", false, null]',
                },
                "attrib_d": {
                    "DataType": "Binary",
                    "BinaryValue": b"bytes value",
                },
            },
        )



import json
import os

import boto3

from awstin.config import aws_config
from awstin.constants import TEST_SNS_ENDPOINT


class SNSTopic:
    """
    A client for typical use of an SNS topic
    """
    def __init__(self, topic_name):
        """
        Parameters
        ----------
        topic_name : str
            Name of the topic
        """
        config = aws_config(endpoint=os.environ.get(TEST_SNS_ENDPOINT))
        self.sns = boto3.resource("sns", **config)
        self.topic = self.sns.create_topic(Name=topic_name)

    def publish(self, message, **attributes):
        """
        Publish a message to the topic

        Parameters
        ----------
        message : str
            The message to send
        **attributes : dict(str, Any)
            Message attributes to add to the message. Value must be castable
            into "String" (str), "String.Array" (list of str),
            "Number" (int or float), and "Binary" (bytes). If it's not
            classified, a bytes cast will be attempted.
        Returns
        -------
        str
            The message's unique ID
        """
        message_attributes = {}
        for key, val in attributes.items():
            if isinstance(val, str):
                message_attributes[key] = {
                    "DataType": "String",
                    "StringValue": val,
                }
            elif isinstance(val, list):
                message_attributes[key] = {
                    "DataType": "String.Array",
                    "StringValue": json.dumps(val),
                }
            elif isinstance(val, int) or isinstance(val, float):
                message_attributes[key] = {
                    "DataType": "Number",
                    "StringValue": str(val),
                }
            else:
                message_attributes[key] = {
                    "DataType": "Binary",
                    "BinaryValue": bytes(val),
                }

        return self.topic.publish(
            Message=message,
            MessageAttributes=message_attributes,
        )["MessageId"]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

description = "Utilities for building and testing AWS applications in Python"

setuptools.setup(
    name="awstin",
    version="0.0.12",
    author="Kevin Duff",
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://https://github.com/k2bd/awstin",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Development Status :: 1 - Planning",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
    ],
    python_requires='>=3.6',
    install_requires=[
        "boto3",
    ],
)



import unittest
from contextlib import ExitStack

import awstin.dynamodb.api as ddb_api
from awstin.dynamodb import DynamoDB
from awstin.dynamodb.testing import temporary_dynamodb_table


class TestDynamoDB(unittest.TestCase):
    def test_dynamodb_table(self):
        with temporary_dynamodb_table("table_name", "hashkey_name") as table:
            test_item = {
                "hashkey_name": "test_value",
                "another_key": 5,
            }

            # Table can be used
            table.put_item(test_item)
            result_item = table["test_value"]
            self.assertEqual(result_item, test_item)

            # DynamoDB can access table
            dynamodb = DynamoDB()
            self.assertEqual(dynamodb["table_name"]["test_value"], test_item)

    def test_dynamodb_table_access_by_shorthand_composite_key(self):
        temp_table = temporary_dynamodb_table(
            "tabname",
            "hashname",
            sortkey_name="sortname",
            sortkey_type="N",
        )
        with temp_table as table:
            test_item = {"hashname": "aaa", "sortname": 555, "aa": "bb"}
            table.put_item(test_item)

            self.assertEqual(table[("aaa", 555)], test_item)

    def test_dynamodb_table_access_by_full_hash_key(self):
        with temporary_dynamodb_table("tabname", "hashname") as table:
            test_item = {"hashname": "123", "nothashname": "456"}
            table.put_item(test_item)

            self.assertEqual(table[{"hashname": "123"}], test_item)

    def test_dynamodb_table_access_by_full_composite_key(self):
        temp_table = temporary_dynamodb_table(
            "tabname",
            "hashname",
            sortkey_name="sortname",
            sortkey_type="N",
        )
        with temp_table as table:
            test_item = {"hashname": "aaa", "sortname": 555, "aa": "bb"}
            table.put_item(test_item)

            self.assertEqual(
                table[{"hashname": "aaa", "sortname": 555}],
                test_item
            )

    def test_dynamodb_table_delete_item(self):
        with temporary_dynamodb_table("tabname", "hashname") as table:
            test_item = {"hashname": "123", "nothashname": "456"}
            table.put_item(test_item)

            self.assertEqual(table["123"], test_item)

            table.delete_item("123")

            with self.assertRaises(KeyError):
                table["123"]

    def test_dynamodb_list_tables(self):
        with temporary_dynamodb_table("test_tab1", "test_key1"):
            with temporary_dynamodb_table("test_tab2", "test_key2"):
                tables = DynamoDB().list_tables()

        self.assertEqual(tables, ["test_tab1", "test_tab2"])

    def test_dynamodb_list_tables_long(self):
        # Reduce page size so test doesn't have to make over 100 tables
        ddb_api._PAGE_SIZE = 5

        dynamodb = DynamoDB()

        table_managers = [
            temporary_dynamodb_table("abc"+str(i), "def"+str(i))
            for i in range(11)
        ]
        with ExitStack() as stack:
            for table in table_managers:
                stack.enter_context(table)

            self.assertEqual(len(dynamodb.list_tables()), 11)

    def test_dynamodb_table_items_generator(self):
        # Page size of DynamoDB item returns is 1MB. Add ~2.2MB of items
        add_items = [
            {"pkey": "item "+str(i), "bigstring": "a" * 100000}
            for i in range(22)
        ]
        with temporary_dynamodb_table("test_tab", "pkey") as table:
            for item in add_items:
                table.put_item(item)

            result_items = list(table.scan())

        self.assertEqual(len(result_items), 22)
        self.assertCountEqual(result_items, add_items)



import unittest
from unittest import mock

from awstin.apigateway.websocket import __name__ as WS_NAME, Websocket


class TestWebsocket(unittest.TestCase):
    def test_websocket_send(self):
        mock_api_client = mock.Mock()
        mock_aws_client = mock.patch(
            WS_NAME + ".boto3.client",
            return_value=mock_api_client,
        )
        with mock_aws_client as m_aws:
            socket = Websocket("endpointurl", stage="dev")
            socket.send("callbackurl", "message")

        m_aws.assert_called_once_with(
            "apigatewaymanagementapi",
            endpoint_url="https://endpointurl/dev"
        )
        mock_api_client.post_to_connection.assert_called_once_with(
            Data="message",
            ConnectionId="callbackurl",
        )



import logging
import unittest

from awstin.awslambda import __name__ as EVENT_NAME
from awstin.awslambda import lambda_handler

EVENT = {
    "requestContext": {"requestId": "an ID"},
}
CONTEXT = {
    "memory_limit_in_mb": 333,
}


class TestEventDecorator(unittest.TestCase):
    def test_custom_event_handler_tuple(self):
        def custom_event_parser(event, context):
            request_id = event["requestContext"]["requestId"]
            memory_limit = context["memory_limit_in_mb"]
            return request_id, memory_limit

        @lambda_handler(custom_event_parser)
        def handle_custom_event(request_id, memory_limit):
            self.assertEqual(request_id, "an ID")
            self.assertEqual(memory_limit, 333)

            return "Result!"

        # Externally the handler takes the normal signature
        with self.assertLogs(EVENT_NAME, logging.INFO) as logs:
            result = handle_custom_event(EVENT, CONTEXT)

        self.assertEqual(result, "Result!")

        self.assertEqual(len(logs.records), 2)
        record1, record2 = logs.records

        self.assertEqual(f"Event: {EVENT!r}", record1.getMessage())
        self.assertEqual("Result: 'Result!'", record2.getMessage())

    def test_custom_event_handler_dict(self):
        def custom_event_parser(event, context):
            request_id = event["requestContext"]["requestId"]
            memory_limit = context["memory_limit_in_mb"]
            return dict(rid=request_id, mem=memory_limit)

        @lambda_handler(custom_event_parser)
        def handle_custom_event(mem, rid):
            self.assertEqual(rid, "an ID")
            self.assertEqual(mem, 333)

            return "Result!"

        result = handle_custom_event(EVENT, CONTEXT)
        self.assertEqual(result, "Result!")

    def test_custom_event_handler_value(self):
        def custom_event_parser(event, context):
            memory_limit = context["memory_limit_in_mb"]
            return memory_limit

        @lambda_handler(custom_event_parser)
        def handle_custom_event(limit):
            self.assertEqual(limit, 333)

            return "Result!"

        result = handle_custom_event(EVENT, CONTEXT)
        self.assertEqual(result, "Result!")
